import torch
import pandas as pd
import numpy as np 
from common.utils import PSNR, cal_ssim, logger_info, _rgb2ycbcr, modcrop
from pathlib import Path
from PIL import Image
import time

# @ray.remote(num_cpus=1, num_gpus=0.3)
def val_image_pair(model, hr_image, lr_image, output_image_path=None, device='cuda'):            
    with torch.inference_mode():
        start_time = time.perf_counter_ns()
        # prepare lr_image
        lr_image = torch.tensor(lr_image).type(torch.float32).permute(2,0,1)
        lr_image = lr_image.unsqueeze(0).to(torch.device(device))
        b, c, h, w = lr_image.shape
        lr_image = lr_image.reshape(b*c, 1, h, w)
        # predict        
        pred_lr_image = model(lr_image)
        # postprocess
        pred_lr_image = pred_lr_image.reshape(b, c, h*model.scale, w*model.scale).squeeze(0).permute(1,2,0).type(torch.uint8)      
        pred_lr_image = pred_lr_image.cpu().numpy()  
        run_time_ns = time.perf_counter_ns() - start_time
        torch.cuda.empty_cache()
        
        if not output_image_path is None:
            Image.fromarray(pred_lr_image).save(output_image_path)

        # metrics
        hr_image = modcrop(hr_image, model.scale)
        left, right = _rgb2ycbcr(pred_lr_image)[:, :, 0], _rgb2ycbcr(hr_image)[:, :, 0]
        return PSNR(left, right, model.scale), cal_ssim(left, right), run_time_ns

def valid_steps(model, datasets, config, log_prefix=""):
    # ray.init(num_cpus=16, num_gpus=1, ignore_reinit_error=True, log_to_driver=False, runtime_env={"working_dir": "../"})
    dataset_names = list(datasets.keys())

    results = []
    for i in range(len(dataset_names)):
        dataset_name = dataset_names[i]
        psnrs, ssims = [], []  
        run_times_ns = []          

        predictions_path = config.valout_dir / dataset_name
        if not predictions_path.exists():
            predictions_path.mkdir()
        
        test_dataset = datasets[dataset_name]
        tasks = []
        for hr_image, lr_image, hr_image_path, lr_image_path in test_dataset:
            output_image_path = predictions_path / f'{Path(hr_image_path).stem}_rcnet.png' if config.save_predictions else None
            task = val_image_pair(model, hr_image, lr_image, output_image_path, device=config.device)
            tasks.append(task)
        # ready_refs, remaining_refs = ray.wait(tasks, num_returns=1, timeout=None)
        # while len(remaining_refs) > 0:
        #     print(f"\rReady {len(ready_refs)+1}/{len(test_dataset)}",  end="     ")
        #     ready_refs, remaining_refs = ray.wait(tasks, num_returns=len(ready_refs)+1, timeout=None)
        # print("\r", end="                                             ")
        # tasks = [ray.get(task) for task in tasks]

        for psnr, ssim, run_time_ns in tasks:
            psnrs.append(psnr)
            ssims.append(ssim)
            run_times_ns.append(run_time_ns)

        config.logger.info(
            '\r{} | Dataset {} | AVG Val PSNR: {:02f}, AVG: SSIM: {:04f}'.format(log_prefix, dataset_name, np.mean(np.asarray(psnrs)), np.mean(np.asarray(ssims))))
        config.writer.flush()

        results.append([
            dataset_name, 
            np.mean(psnrs), 
            np.mean(ssims), 
            np.mean(run_times_ns)*1e-9, 
            np.percentile(run_times_ns, q=95)*1e-9])

    results = pd.DataFrame(results, columns=['Dataset', 'PSNR', 'SSIM', f'AVG {config.device} Time, s', f'P95 {config.device} Time, s']).set_index('Dataset')

    return results